﻿const int IterationCount = 10_000_000;

Console.WriteLine($"{Environment.Version} | {System.Runtime.InteropServices.RuntimeInformation.FrameworkDescription} | {System.Runtime.InteropServices.RuntimeInformation.ProcessArchitecture} | {System.Runtime.InteropServices.RuntimeInformation.OSDescription}");

Console.WriteLine("+++ BEFORE everything:");
PrintGCHeapInfo();

var context = new MemoryLeakingBenchmarks();
context.GlobalSetup();

Console.WriteLine("+++ AFTER .GlobalSetup() call:");
PrintGCHeapInfo();

long total = 0;

//

Console.WriteLine(">>> Allocating class instances + List >>>");
Console.WriteLine();

for (int i = 0; i < IterationCount; i++) {
	total += context.ClassAllocSumForget();
}
Console.WriteLine($"+++ AFTER {nameof(MemoryLeakingBenchmarks.ClassAllocSumForget)}:");
PrintGCHeapInfo();

for (int i = 0; i < IterationCount; i++) {
	total += context.ClassAllocSumSaveInLastField();
}
Console.WriteLine($"+++ AFTER {nameof(MemoryLeakingBenchmarks.ClassAllocSumSaveInLastField)}:");
PrintGCHeapInfo();

for (int i = 0; i < IterationCount; i++) {
	total += context.ClassAllocSumSaveInSmallArray();
}
Console.WriteLine($"+++ AFTER {nameof(MemoryLeakingBenchmarks.ClassAllocSumSaveInSmallArray)}:");
PrintGCHeapInfo();

for (int i = 0; i < IterationCount; i++) {
	total += context.ClassAllocSumSaveInMediumArray();
}
Console.WriteLine($"+++ AFTER {nameof(MemoryLeakingBenchmarks.ClassAllocSumSaveInMediumArray)}:");
PrintGCHeapInfo();

for (int i = 0; i < IterationCount; i++) {
	total += context.ClassAllocSumSaveInBigArray();
}
Console.WriteLine($"+++ AFTER {nameof(MemoryLeakingBenchmarks.ClassAllocSumSaveInBigArray)}:");
PrintGCHeapInfo();

//

Console.WriteLine(">>> Allocating List with embedded struct instances >>>");
Console.WriteLine();

for (int i = 0; i < IterationCount; i++) {
	total += context.StructAllocSumForget();
}
Console.WriteLine($"+++ AFTER {nameof(MemoryLeakingBenchmarks.StructAllocSumForget)}:");
PrintGCHeapInfo();

for (int i = 0; i < IterationCount; i++) {
	total += context.StructAllocSumSaveInLastField();
}
Console.WriteLine($"+++ AFTER {nameof(MemoryLeakingBenchmarks.StructAllocSumSaveInLastField)}:");
PrintGCHeapInfo();

for (int i = 0; i < IterationCount; i++) {
	total += context.StructAllocSumSaveInMediumArray();
}
Console.WriteLine($"+++ AFTER {nameof(MemoryLeakingBenchmarks.StructAllocSumSaveInMediumArray)}:");
PrintGCHeapInfo();

for (int i = 0; i < IterationCount; i++) {
	total += context.StructReuseSumClear();
}
Console.WriteLine($"+++ AFTER {nameof(MemoryLeakingBenchmarks.StructReuseSumClear)}:");
PrintGCHeapInfo();


static void PrintGCHeapInfo() {
	for (int i = 0; i <= GC.MaxGeneration; i++) {
		Console.WriteLine($"Gen {i} collection count: {GC.CollectionCount(i)}");
	}

    Console.WriteLine("Heap sizes:");
    var memoryInfo = GC.GetGCMemoryInfo();
	for (int i = 0; i <= GC.MaxGeneration + 1; i++) {
		if (i > GC.MaxGeneration) {
			Console.WriteLine("  LOH (Large Object Heap):");
		} else {
			Console.WriteLine($"  Gen {i} in SOH (Small Object Heap):");
		}
		var genInfo = memoryInfo.GenerationInfo[i];
		Console.WriteLine($"    + Size before last GC : {genInfo.SizeBeforeBytes:N0} B");
		Console.WriteLine($"    + Size after last GC  : {genInfo.SizeAfterBytes:N0} B");
	}
	Console.WriteLine();
}

public class A {
	public int X;
}

public struct S {
	public int X;
}

public class MemoryLeakingBenchmarks {
	public A? LastA;

	public A[]? SmallArray;
	public A[]? MediumArray;
	public A[]? BigArray;

	public S LastS = default(S);

	public S[]? MediumArrayOfStructs;

	private long _iterationIndex;

	public void GlobalSetup() {
		SmallArray = new A[1024];                           // SOH <-       8 kB <- 64-bit arch: 8B per A reference * array of 1024 items + overhead (ignored and rounded for simplicity)
		MediumArray = new A[16 * SmallArray.Length];        // LOH <-     128 kB 
		BigArray = new A[MediumArray.Length * 1024 * 4];    // LOH <- 524 288 kB = 512 MB 

		MediumArrayOfStructs = new S[MediumArray.Length];   // SOH <-      64 kB <- 1 S instance = 1 Int32 field = 4 B * array of 16 * 1024 items + overhead (ignored and rounded for simplicity)

		_iterationIndex = 0;
	}

	private const int InstancePerIterationCount = 8;

	private List<A> CreateNewListOfA() {
		var list = new List<A>();
		for (int i = 0; i < InstancePerIterationCount; i++) {
			list.Add(new A { X = i });
		}
		return list;
	}

	public int ClassAllocSumForget() {
		List<A> list = CreateNewListOfA();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
		}
		return sum;
	}

	public int ClassAllocSumSaveInLastField() {
		List<A> list = CreateNewListOfA();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
			LastA = list[i];
		}
		return sum;
	}

	public int ClassAllocSumSaveInSmallArray() {
		List<A> list = CreateNewListOfA();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
			SmallArray![(_iterationIndex++) % SmallArray.Length] = list[i];
		}
		return sum;
	}

	public int ClassAllocSumSaveInMediumArray() {
		List<A> list = CreateNewListOfA();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
			MediumArray![(_iterationIndex++) % MediumArray.Length] = list[i];
		}
		return sum;
	}

	public int ClassAllocSumSaveInBigArray() {
		List<A> list = CreateNewListOfA();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
			BigArray![(_iterationIndex++) % BigArray.Length] = list[i];
		}
		return sum;
	}

	//

	private List<S> CreateNewListOfS() {
		var list = new List<S>();
		for (int i = 0; i < InstancePerIterationCount; i++) {
			list.Add(new S { X = i });
		}
		return list;
	}

	public int StructAllocSumForget() {
		List<S> list = CreateNewListOfS();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
		}
		return sum;
	}

	public int StructAllocSumSaveInLastField() {
		List<S> list = CreateNewListOfS();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
			LastS = list[i];
		}
		return sum;
	}

	public int StructAllocSumSaveInMediumArray() {
		List<S> list = CreateNewListOfS();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
			MediumArrayOfStructs![(_iterationIndex++) % MediumArrayOfStructs.Length] = list[i];
		}
		return sum;
	}

	private List<S> _reusableSList = new List<S>();

	private List<S> ReuseListOfS() {
		_reusableSList.Clear();
		var list = _reusableSList;
		for (int i = 0; i < InstancePerIterationCount; i++) {
			list.Add(new S { X = i });
		}
		return list;
	}

	public int StructReuseSumClear() {
		List<S> list = ReuseListOfS();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
		}
		return sum;
	}
}
