﻿using BenchmarkDotNet.Attributes;
using BenchmarkDotNet.Running;

BenchmarkRunner.Run<MemoryLeakingBenchmarks>();

public class A {
	public int X;
}

public struct S {
	public int X;
}

[MemoryDiagnoser]
public class MemoryLeakingBenchmarks {
	public A? LastA;

	public A[]? SmallArray;
	public A[]? MediumArray;
	public A[]? BigArray;

	public S LastS = default(S);

	public S[]? MediumArrayOfStructs;

	private long _iterationIndex;

	[GlobalSetup]
	public void GlobalSetup() {
		SmallArray = new A[1024];                           // SOH <-       8 kB <- 64-bit arch: 8B per A reference * array of 1024 items + overhead (ignored and rounded for simplicity)
		MediumArray = new A[16 * SmallArray.Length];        // LOH <-     128 kB 
		BigArray = new A[MediumArray.Length * 1024 * 4];    // LOH <- 524 288 kB = 512 MB 

		MediumArrayOfStructs = new S[MediumArray.Length];   // SOH <-      64 kB <- 1 S instance = 1 Int32 field = 4 B * array of 16 * 1024 items + overhead (ignored and rounded for simplicity)

		_iterationIndex = 0;
	}

	private const int InstancePerIterationCount = 8;

	//

	private List<A> CreateNewListOfA() {
		var list = new List<A>();
		for (int i = 0; i < InstancePerIterationCount; i++) {
			list.Add(new A { X = i });
		}
		return list;
	}

	[Benchmark]
	public int ClassAllocSumForget() {
		List<A> list = CreateNewListOfA();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X + /* for fairness: */ (int) (_iterationIndex++ % MediumArray!.Length);
		}
		return sum;
	}

	[Benchmark]
	public int ClassAllocSumSaveInLastField() {
		List<A> list = CreateNewListOfA();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X + /* for fairness: */ (int) (_iterationIndex++ % MediumArray!.Length);
			LastA = list[i];
		}
		return sum;
	}

	[Benchmark]
	public int ClassAllocSumSaveInSmallArray() {
		List<A> list = CreateNewListOfA();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
			SmallArray![_iterationIndex++ % SmallArray.Length] = list[i];
		}
		return sum;
	}

	[Benchmark]
	public int ClassAllocSumSaveInMediumArray() {
		List<A> list = CreateNewListOfA();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
			MediumArray![_iterationIndex++ % MediumArray.Length] = list[i];
		}
		return sum;
	}
	
	[Benchmark]
	public int ClassAllocSumSaveInBigArray() {
		List<A> list = CreateNewListOfA();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
			BigArray![_iterationIndex++ % BigArray.Length] = list[i];
		}
		return sum;
	}

	//

	private List<S> CreateNewListOfS() {
		var list = new List<S>();
		for (int i = 0; i < InstancePerIterationCount; i++) {
			list.Add(new S { X = i });
		}
		return list;
	}

	[Benchmark]
	public int StructAllocSumForget() {
		List<S> list = CreateNewListOfS();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X + /* for fairness: */ (int) (_iterationIndex++ % MediumArray!.Length);
		}
		return sum;
	}

	[Benchmark]
	public int StructAllocSumSaveInLastField() {
		List<S> list = CreateNewListOfS();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X + /* for fairness: */ (int) (_iterationIndex++ % MediumArray!.Length);
			LastS = list[i];
		}
		return sum;
	}

	[Benchmark]
	public int StructAllocSumSaveInMediumArray() {
		List<S> list = CreateNewListOfS();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
			MediumArrayOfStructs![_iterationIndex++ % MediumArrayOfStructs.Length] = list[i];
		}
		return sum;
	}

	private List<S> _reusableSList = new List<S>();

	private List<S> ReuseListOfS() {
		_reusableSList.Clear();
		var list = _reusableSList;
		for (int i = 0; i < InstancePerIterationCount; i++) {
			list.Add(new S { X = i });
		}
		return list;
	}

	[Benchmark]
	public int StructReuseSumClear() {
		List<S> list = ReuseListOfS();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X + /* for fairness: */ (int) (_iterationIndex++ % MediumArray!.Length);
		}
		return sum;
	}

	// Just for completeness - unfair version, does not do the expensive % calculation, etc.:

	[Benchmark]
	public int StructAllocSumForget_UNFAIR() {
		List<S> list = CreateNewListOfS();

		int sum = 0;
		for (int i = 0; i < list.Count; i++) {
			sum += list[i].X;
		}
		return sum;
	}
}
